#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from argilla_v1.client.feedback.schemas.response_values import (
    ResponseValue,
    normalize_response_value,
)
from argilla_v1.pydantic_v1 import BaseModel, Extra, confloat, validator


class SuggestionSchema(BaseModel):
    """Schema for the suggestions for the questions related to the record.

    Args:
        question_name: name of the question in the `FeedbackDataset`.
        type: type of the question. Defaults to None. Possible values are `model` or `human`.
        score: score of the suggestion. Defaults to None.
        value: value of the suggestion, which should match the type of the question.
        agent: agent that generated the suggestion. Defaults to None.

    Examples:
        >>> from argilla_v1.client.feedback.schemas.suggestions import SuggestionSchema
        >>> SuggestionSchema(
        ...     question_name="question-1",
        ...     type="model",
        ...     score=0.9,
        ...     value="This is the first suggestion",
        ...     agent="agent-1",
        ... )
    """

    question_name: str
    value: ResponseValue
    score: Union[Optional[confloat(ge=0, le=1)], List[confloat(ge=0, le=1)]] = None
    type: Optional[Literal["model", "human"]] = None
    agent: Optional[str] = None

    _normalize_value = validator("value", allow_reuse=True, always=True)(normalize_response_value)

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    @validator("score")
    @classmethod
    def validate_score(cls, score, values):
        value = values["value"]

        if isinstance(value, list) and isinstance(score, list) and len(value) != len(score):
            raise ValueError("The length of the score list should match the length of the value list")

        return score

    def to_server_payload(self, question_name_to_id: Dict[str, UUID]) -> Dict[str, Any]:
        """Method that will be used to create the payload that will be sent to Argilla
        to create a `SuggestionSchema` for a `FeedbackRecord`."""
        payload = self.dict(exclude_unset=True, include={"type", "score", "value", "agent"})
        payload["question_id"] = str(question_name_to_id[self.question_name])

        return payload



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import random
import warnings
from collections import defaultdict
from typing import Any, Dict, List, Union

from rich.progress import Progress

from argilla_v1.client.users import User
from argilla_v1.client.workspaces import Workspace


def check_user(user_to_check: Union[str, User]) -> User:
    """
    Helper function to check if the input is a User object. If it's a string, it attempts to retrieve the User object.
    If the User does not exist, it creates a new User with a default password and role.

    Args:
        user_to_check: a user object or a string that represents a username

    Returns:
        The User object corresponding to the input.
    """
    if isinstance(user_to_check, User):
        user = user_to_check
    else:
        try:
            user = User.from_name(user_to_check)
        except ValueError:
            user = User.create(username=user_to_check, first_name=user_to_check, password="12345678", role="annotator")
            warnings.warn(
                f"The user {user.username} was created with a default password. We recommend you to change it for security reasons.",
                UserWarning,
            )
    return user


def check_workspace(workspace_to_check: str) -> Workspace:
    """
    Helper function to check if the workspace exists. If it does not exist, it creates a new one.

    Args:
        workspace_to_check: a workspace string name

    Returns:
        The Workspace object corresponding to the input.
    """
    try:
        workspace = Workspace.from_name(workspace_to_check)
    except:
        workspace = Workspace.create(workspace_to_check)
    return workspace


def assign_records_to_groups(
    groups: Dict[str, List[Any]], records: List[Any], overlap: int, shuffle: bool = True
) -> Dict[str, Dict[str, Any]]:
    """
    Assign records to predefined groups with controlled overlap (for the groups) and optional shuffle. All members of the same group will annotate the same records.

    Args:
        groups: A dictionary where keys are group names and values are lists of users names or objects.
        records: A list of records to be assigned.
        overlap: The number of times each record is assigned to consecutive groups (0 for no overlap).
        shuffle: If True, shuffle the records before assignment. Defaults to True.

    Returns:
        A dictionary where each key is a group and its value is another dictionary, which maps usernames to their respective assigned records.

    Raises:
        ValueError: If `overlap` is higher than the number of groups or negative.
    """
    if overlap < 0 or overlap >= len(groups.keys()):
        raise ValueError("Overlap must be less than the number of groups and must not be negative.")

    if len(records) < len(groups.keys()):
        warnings.warn(
            "The number of groups is higher than the number of records. Some users will not be assigned any records.",
            UserWarning,
        )

    if shuffle:
        random.shuffle(records)

    assignments = {}
    assignments_grouped = {}
    group_names = list(groups.keys())
    num_groups = len(group_names)
    overlap = 1 if overlap == 0 else overlap

    group_records = defaultdict(list)
    with Progress() as progress:
        task = progress.add_task("[green]Processing records...", total=len(records))

        for idx, record in enumerate(records):
            for offset in range(overlap):
                group_index = (idx + offset) % num_groups
                group_name = group_names[group_index]
                group_records[group_name].append(record)

            progress.update(task, advance=1)

    for group, users in groups.items():
        users = [check_user(user) for user in users]
        for user in users:
            assignments[user] = group_records[group]

        assignments_grouped[group] = {user.username: assignments.get(user, []) for user in users}

    return assignments_grouped


def assign_records_to_individuals(
    users: List[Any], records: List[Any], overlap: int, shuffle: bool = True
) -> Dict[str, List[Any]]:
    """
    Assign records to users with controlled overlap and optional shuffle.

    Args:
        users: A list of user objects, each with a 'username' attribute.
        records: A list of record objects to be assigned to users.
        overlap: The number of times each record is assigned to consecutive users (0 for no overlap).
        shuffle: If True, the records list will be shuffled before assignment. Defaults to True.

    Returns:
        A dictionary where keys are usernames and values are lists of assigned records.

    Raises:
        ValueError: If `overlap` is higher than the number of users or negative.
    """
    if overlap < 0 or overlap >= len(users):
        raise ValueError("Overlap must be less than the number of users and must not be negative.")

    if len(records) < len(users):
        warnings.warn(
            "The number of users is higher than the number of records. Some users will not be assigned any records.",
            UserWarning,
        )

    if shuffle:
        random.shuffle(records)

    users = [check_user(user) for user in users]
    assignments = {user.username: [] for user in users}

    num_users = len(users)
    overlap = 1 if overlap == 0 else overlap

    with Progress() as progress:
        task = progress.add_task("[green]Processing records...", total=len(records))

        for idx, record in enumerate(records):
            for offset in range(overlap):
                user_index = (idx + offset) % num_users
                user = users[user_index].username
                assignments[user].append(record)

            progress.update(task, advance=1)

    return assignments


def assign_records(
    users: Union[Dict[str, List[Any]], List[Any]], records: List[Any], overlap: int, shuffle: bool = True
) -> Union[Dict[str, List[Any]], Dict[str, Dict[str, Any]]]:
    """
    Assign records to either groups or individuals, with controlled overlap and optional shuffle.

    Args:
        users: Either a dictionary of groups or a list of individual user objects.
        records: A list of record objects to be assigned.
        overlap: The number of times each record is assigned to consecutive users or groups (0 for no overlap).
        shuffle: If True, the records list will be shuffled before assignment. Defaults to True.

    Returns:
        A dictionary where each key is a group and its value is another dictionary, which maps usernames to their respective assigned records.
        Or a dictionary where keys are usernames and values are lists of assigned records.

    Examples:
        >>> from argilla_v1.client.feedback.utils import assign_records
        >>> individual_assignments = assign_records([user1, user2, user3], records, 0, False)
        >>> group_assignments = assign_records({group1: [user1, user2], group2: [user3]}, records, 1, False)

    """
    if isinstance(users, dict):
        return assign_records_to_groups(users, records, overlap, shuffle)
    elif isinstance(users, list):
        return assign_records_to_individuals(users, records, overlap, shuffle)


def assign_workspaces(
    assignments: Union[Dict[str, List[Any]], Dict[str, Dict[str, Any]]], workspace_type: str
) -> Dict[str, List[Any]]:
    """
    Assign workspaces (and create them if needed) to either groups or individuals.

    Args:
        assignments: Either a dictionary of groups or a dictionary of users.
        workspace_type: Either 'group' (each group in a workspace), 'group_personal' (each member in a workspace) or 'individual' (each person in a workspace).

    Returns:
        A dictionary where each key is a workspace name and its value is a list of user names.

    Examples:
        >>> from argilla_v1.client.feedback.utils import assign_workspaces
        >>> wk_assignments = assign_workspaces(group_assignments, "group")
        >>> wk_assignments = assign_workspaces(group_assignments, "group_personal")
        >>> wk_assignments = assign_workspaces(individual_assignments, "individual")

    """
    wk_assignments = {}

    for group, users in assignments.items():
        if workspace_type == "group":
            workspace_name = group
            user_ids = [check_user(user).id for user in users.keys()]

        elif workspace_type == "group_personal":
            for user in users.keys():
                workspace_name = user
                user_ids = [check_user(user).id]
                workspace = check_workspace(workspace_name)

                for user_id in user_ids:
                    try:
                        workspace.add_user(user_id)
                    except:
                        pass

                wk_assignments[workspace_name] = [User.from_id(user.id).username for user in workspace.users]

            continue

        elif workspace_type == "individual":
            workspace_name = group
            user_ids = [check_user(group).id]

        workspace = check_workspace(workspace_name)

        for user_id in user_ids:
            try:
                workspace.add_user(user_id)
            except:
                pass

        wk_assignments[workspace_name] = [User.from_id(user.id).username for user in workspace.users]

    return wk_assignments



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from argilla_v1 import TokenClassificationRecord
from argilla_v1.client.api import Api
from argilla_v1.monitoring.base import BaseMonitor
from argilla_v1.monitoring.types import MissingType

try:
    from flair import __version__ as _flair_version
    from flair.data import Sentence
    from flair.models import SequenceTagger
except (ImportError, ModuleNotFoundError):
    _flair_version = None
    Sentence = MissingType
    SequenceTagger = MissingType


class FlairMonitor(BaseMonitor):
    def _prepare_log_data(
        self,
        data: List[Tuple[Sentence, Dict[str, Any]]],
    ) -> Dict[str, Any]:
        return dict(
            records=[
                TokenClassificationRecord(
                    text=sentence.to_original_text(),
                    tokens=[token.text for token in sentence.tokens],
                    metadata=meta,
                    prediction_agent=self.agent,
                    event_timestamp=datetime.utcnow(),
                    prediction=[
                        (
                            label.value,
                            span.start_position,
                            span.end_position,
                            label.score,
                        )
                        for label, span in zip(
                            sentence.get_labels(self.__model__.tag_type), sentence.get_spans(self.__model__.tag_type)
                        )
                    ],
                )
                for sentence, meta in data
            ],
            name=self.dataset,
            tags={**(self.tags or {}), "flair_version": _flair_version},
        )

    def predict(self, sentences: Union[List[Sentence], Sentence], *args, **kwargs):
        metadata = kwargs.pop("metadata", None)
        result = self.__model__.predict(sentences, *args, **kwargs)

        if not isinstance(sentences, list):
            sentences = [sentences]

        if not metadata:
            metadata = [{}] * len(sentences)

        filtered_data = [(sentence, meta) for sentence, meta in zip(sentences, metadata) if self.is_record_accepted()]
        if filtered_data:
            self._log_future = self.send_records(filtered_data)

        return result


def flair_monitor(
    pl: SequenceTagger, api: Api, dataset: str, sample_rate: float, log_interval: float
) -> Optional[SequenceTagger]:
    return FlairMonitor(
        pl,
        api=api,
        dataset=dataset,
        sample_rate=sample_rate,
        log_interval=log_interval,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from types import TracebackType
from typing import Any, ContextManager, Dict, Optional, Type

from packaging.version import parse

from argilla_v1.client.apis import AbstractApi
from argilla_v1.client.sdk.client import AuthenticatedClient
from argilla_v1.client.sdk.commons.errors import ApiCompatibilityError
from argilla_v1.pydantic_v1 import BaseModel


class ApiInfo(BaseModel):
    version: Optional[str] = None


class ElasticSearchVersionDetails(BaseModel):
    distribution: Optional[str] = None
    number: str
    build_flavor: Optional[str] = None
    build_type: str
    build_hash: str
    build_date: str
    build_snapshot: bool
    lucene_version: str
    minimum_wire_compatibility_version: str
    minimum_index_compatibility_version: str


class ElasticSearch(BaseModel):
    name: str
    cluster_name: str
    cluster_uuid: str
    version: ElasticSearchVersionDetails
    tagline: str


class ApiStatus(BaseModel):
    version: str
    elasticsearch: ElasticSearch
    mem_info: Dict[str, Any]


class Status(AbstractApi):
    def get_info(self) -> ApiInfo:
        """
        Get the connected API info

        Returns:
            ApiInfo: The API info
        """

        response = self.http_client.get("/api/_info")
        return ApiInfo.parse_obj(response)

    def get_status(self) -> ApiStatus:
        response = self.http_client.get("/api/_status")
        return ApiStatus.parse_obj(response)


class _ApiCompatibilityContextManager(ContextManager):
    def __init__(
        self,
        client: AuthenticatedClient,
        min_version: str,
    ):
        self._min_version = parse(min_version)
        self._status_api = Status(client=client) if client else None

    def __enter__(self):
        if self._status_api:
            api_info = self._status_api.get_info()
            api_version = api_info.version

            api_version = parse(api_version)
            if api_version.is_devrelease:
                api_version = parse(api_version.base_version)
            if not api_version >= self._min_version:
                raise ApiCompatibilityError(str(self._min_version), api_version=api_version)
        pass

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        pass


def api_compatibility(api: AbstractApi, min_version: str):
    """
    Handles problems related to server API compatibility.

    Args:
        api: The api component where apply the api compatibility
        min_version: the minimal version that the argilla client can work with
    """
    return _ApiCompatibilityContextManager(
        min_version=min_version,
        client=api.http_client,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from typing import List, Optional, Union
from uuid import uuid4

from datasets import DatasetDict

from argilla_v1.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla_v1.training.base import ArgillaTrainerSkeleton
from argilla_v1.utils.dependency import require_dependencies


class AutoTrainMixin:
    def prepare_dataset(self, data_dict: Optional[dict] = {}) -> None:
        """
        This function prepares a dataset for autotrain using a dictionary of data and specific column
        mappings.

        Args:
          data_dict_params: A dictionary that contains additional parameters to be passed
            to the `AutoTrainDataset` constructor. These parameters are optional and can be used to
            customize the dataset preparation process.
        """
        from autotrain.dataset import AutoTrainDataset

        self.dset = AutoTrainDataset(
            task=self.task,
            token=self.HF_TOKEN,
            username=self.AUTOTRAIN_USERNAME,
            project_name=self.project_name,
            **data_dict,
            percent_valid=None,
            column_mapping={
                "text": "text",
                "label": "label",
            },
        )

    def get_project_cost(self):
        """
        This function returns the project cost based on various parameters.

        Returns:
          The `get_project_cost` function is being called with several arguments and its return value is
        being returned by the `get_project_cost` method. The exact return value depends on the
        implementation of the `get_project_cost` function.
        """
        from autotrain.utils import get_project_cost

        return get_project_cost(
            username=self.AUTOTRAIN_USERNAME,
            token=self.HF_TOKEN,
            task=self.task,
            num_samples=self._num_samples,
            num_models=self.trainer_kwargs["autotrain"][0]["num_models"]
            if self._model.lower() == "autotrain"
            else len(self.trainer_kwargs["hub_model"]),
        )

    def initialize_project(self):
        """
        This function initializes a project with a dataset, hub model, and job parameters, and logs
        information about the project and its cost.
        """
        import copy

        from autotrain.project import Project

        self.project = Project(
            dataset=self.dset,
            hub_model=None if self._model.lower() == "autotrain" else self._model,
            job_params=copy.deepcopy(self.get_job_params()),
        )
        self._logger.info(self.project)
        self._logger.info(f"Project cost: {self.get_project_cost()}")

    def get_job_params(self):
        if self._model.lower() == "autotrain":
            job_params = self.trainer_kwargs["autotrain"]
            model_choice = "autotrain"
        else:
            job_params = self.trainer_kwargs["hub_model"]
            model_choice = "HuggingFace Hub"
        if model_choice == "autotrain":
            if len(job_params) > 1:
                raise ValueError("Only one job parameter is allowed for AutoTrain.")
            job_params[0].update({"task": self.task})
        elif model_choice == "HuggingFace Hub":
            for i in range(len(job_params)):
                job_params[i].update({"task": self.task})
        return job_params


class ArgillaAutoTrainTrainer(ArgillaTrainerSkeleton, AutoTrainMixin):
    _logger = logging.getLogger("ArgillaAutoTrainTrainer")
    _logger.setLevel(logging.INFO)

    try:
        AUTOTRAIN_USERNAME = os.environ["AUTOTRAIN_USERNAME"]
        HF_TOKEN = os.environ["HF_AUTH_TOKEN"]
    except KeyError:
        raise KeyError("Please set the `AUTOTRAIN_USERNAME` and `HF_AUTH_TOKEN` environment variables.")

    require_dependencies(["autotrain-advanced", "datasets"])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.project_name = f"{self._workspace}_{self._name}_{str(uuid4())[:8]}"

        if self._seed:
            self._logger.warning("Setting a seed is not supported by `autotrain-advanced`.")
            self._seed = 42

        if self._model is None:
            self._model = "bert-base-uncased"
            self._logger.warning(f"No model defined. Using the default model {self._model}.")

        data_dict = {}
        self._num_samples = 0
        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
            data_dict["valid_data"] = [self._eval_dataset.to_pandas()]
            self._num_samples += len(self._eval_dataset)
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None
            data_dict["valid_data"] = []
        data_dict["train_data"] = [self._train_dataset.to_pandas()]
        self._num_samples += len(self._train_dataset)

        if self._record_class == TextClassificationRecord:
            if self._multi_label:
                raise NotImplementedError(
                    "TextClassificaiton `multi_label=True` is not supported by `autotrain-advanced`."
                )
            elif self._multi_label is False:
                n_classes = len(self._train_dataset.features["label"].names)
                self.task = "text_multi_class_classification" if n_classes > 2 else "text_binary_classification"
                self._id2label = dict(enumerate(self._train_dataset.features["label"].names))
                self._label_list = self._train_dataset.features["label"].names
            self._label2id = {v: k for k, v in self._id2label.items()}
        elif self._record_class == TokenClassificationRecord:
            raise NotImplementedError(
                "`Text2Text` and `TokenClassification` tasks are not supported by `autotrain-advanced`."
            )

        self.init_training_args()

        self.prepare_dataset(data_dict=data_dict)
        self.initialize_project()

    def init_training_args(self):
        from autotrain.params import Params

        if not hasattr(self, "trainer_kwargs"):
            self.trainer_kwargs = {"autotrain": [{}], "hub_model": [{}]}
        for training_type in ["autotrain", "hub_model"]:
            for idx, set_params in enumerate(self.trainer_kwargs[training_type]):
                params = Params(
                    task=self.task,
                    training_type=training_type,
                ).get()
                for key, value in params.items():
                    if key not in set_params:
                        set_params[key] = value.DEFAULT

                self.trainer_kwargs[training_type][idx] = set_params

    def update_config(self, **kwargs):
        """
        Updates the `setfit_model_kwargs` and `setfit_trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        self._model = kwargs.get("model", self._model)
        if self._model.lower() == "autotrain":
            self._model = "autotrain"
        if "hub_model" in kwargs:
            if not isinstance(kwargs["hub_model"], list):
                raise ValueError("hub_model must be a list of dictionaries.")
            self.trainer_kwargs["hub_model"] = kwargs["hub_model"]
        if "autotrain" in kwargs:
            if not isinstance(kwargs["autotrain"], list):
                raise ValueError("autotrain must be a list of dictionaries.")
            self.trainer_kwargs["autotrain"] = kwargs["autotrain"]

        self.init_training_args()
        self.initialize_project()

    def __repr__(self):
        formatted_string = [
            "Choose EITHER `autotrain` OR a model from the hub `hub_model` as main key to update parameters.",
            f"model: {self._model}",
        ]
        for arg_dict_key, arg_dict_single in self.trainer_kwargs.items():
            arg_dict_key += ": List[dict]"
            formatted_string.append(arg_dict_key)
            for idx, item in enumerate(arg_dict_single):
                for key, val in item.items():
                    formatted_string.append(f"\tjob{idx+1}-{key}: {val}")
        return "\n".join(formatted_string)

    def train(self, output_dir: str):
        """
        We create a SetFitModel object from a pretrained model, then create a SetFitTrainer object with
        the model, and then train the model
        """
        self.project.dataset.prepare()
        project_id = self.project.create()
        self.project.approve(project_id)

    def init_model(self, new: bool = False):
        pass

    def init_pipeline(self):
        pass

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        self._logger.error(
            f"Use ArgillaTrainer(name={self._name}, workspace={self._workspace}, framework='transformers', model='my_model_name') for inference via `.predict()`."
        )

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self._logger.warning("Models are saved on the HuggingFace Hub, so this function is not supported.")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .datasets import app as datasets_app
from .info import app as info_app
from .login import app as login_app
from .logout import app as logout_app
from .training import app as training_app
from .users import app as users_app
from .whoami import app as whoami_app
from .workspaces import app as workspaces_app



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .commons import records_status, text_length
from .text_classification import f1, f1_multilabel
from .token_classification import (
    entity_capitalness,
    entity_consistency,
    entity_density,
    entity_labels,
    mention_length,
    token_capitalness,
    token_frequency,
    token_length,
    tokens_length,
)
from .token_classification import f1 as ner_f1

__all__ = [
    text_length,
    records_status,
    f1,
    f1_multilabel,
    entity_capitalness,
    entity_consistency,
    entity_density,
    entity_labels,
    ner_f1,
    mention_length,
    token_capitalness,
    token_frequency,
    token_length,
    tokens_length,
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import dataclasses
import functools
import logging
import threading
import time
from typing import Any, Dict, List, Optional

import schedule

import argilla_v1
from argilla_v1.client import singleton
from argilla_v1.client.sdk.commons.errors import NotFoundApiError
from argilla_v1.listeners.models import ListenerAction, ListenerCondition, Metrics, RGListenerContext, Search


@dataclasses.dataclass
class RGDatasetListener:
    """
    The argilla dataset listener class

    Args:
        dataset: The dataset over which listener is created
        action: The action to execute when the condition is satisfied
        metrics: A list of metrics ids that will be required in condition
        query: The query string to apply
        query_params: Defined parameters used dynamically in the provided query
        condition: The condition to satisfy to execute the action
        query_records: If ``False``, the records won't be passed as argument to the action.
            Default: ``True``
        interval_in_seconds: How often the listener is executed. Default to 30 seconds
    """

    _LOGGER = logging.getLogger(__name__)

    dataset: str
    action: ListenerAction
    metrics: Optional[List[str]] = None
    query: Optional[str] = None
    query_params: Optional[Dict[str, Any]] = None
    condition: Optional[ListenerCondition] = None
    query_records: bool = True
    interval_in_seconds: int = 30

    @property
    def formatted_query(self) -> Optional[str]:
        """Formatted query using defined query params, if any"""
        if self.query is None:
            return None
        return self.query.format(**(self.query_params or {}))

    __listener_job__: Optional[schedule.Job] = dataclasses.field(init=False, default=None)
    __stop_schedule_event__ = None
    __current_thread__ = None
    __scheduler__ = schedule.Scheduler()

    def __post_init__(self):
        self.metrics = self.metrics or []
        self._validate()

    def _validate(self):
        try:
            query = self.formatted_query
            if query:
                self._LOGGER.debug(f"Initial listener query {query}")
        except KeyError as kex:
            raise KeyError("Missing query parameter:", kex)

    def is_running(self):
        """True if listener is running"""
        return self.__listener_job__ is not None

    def __catch_exceptions__(self, cancel_on_failure=False):
        def catch_exceptions_decorator(job_func):
            @functools.wraps(job_func)
            def wrapper(*args, **kwargs):
                try:
                    return job_func(*args, **kwargs)
                except:  # noqa: E722
                    import traceback

                    print(traceback.format_exc())
                    if cancel_on_failure:
                        self.stop()  # We stop the scheduler

            return wrapper

        return catch_exceptions_decorator

    def start(self, *action_args, **action_kwargs):
        """
        Start listen to changes in the dataset. Additionally, args and kwargs can be passed to action
        by using the `action_*` arguments

        If the listener is already started, a ``ValueError`` will be raised

        """
        if self.is_running():
            raise ValueError("Listener is already running")

        job_step = self.__catch_exceptions__(cancel_on_failure=True)(self.__listener_iteration_job__)

        self.__listener_job__ = self.__scheduler__.every(self.interval_in_seconds).seconds.do(
            job_step, *action_args, **action_kwargs
        )

        class _ScheduleThread(threading.Thread):
            _WAIT_EVENT = threading.Event()

            _THREAD_LOGGER = logging.getLogger(__name__)

            @classmethod
            def run(cls):
                cls._THREAD_LOGGER.debug("Running listener thread...")
                while not cls._WAIT_EVENT.is_set():
                    self.__scheduler__.run_pending()
                    time.sleep(self.interval_in_seconds - 1)
                cls._THREAD_LOGGER.debug("Stopping listener thread...")

            @classmethod
            def stop(cls):
                cls._WAIT_EVENT.set()

        self.__current_thread__ = _ScheduleThread()
        self.__current_thread__.start()

    def stop(self):
        """
        Stops listener if it's still running.

        If listener is already stopped, a ``ValueError`` will be raised

        """
        if not self.is_running():
            raise ValueError("Listener is not running")

        self.__scheduler__.cancel_job(self.__listener_job__)
        self.__listener_job__ = None
        self.__current_thread__.stop()
        self.__current_thread__.join()  # TODO: improve it!

    def __listener_iteration_job__(self, *args, **kwargs):
        """
        Execute a complete listener iteration. The iteration consists on:

        1. Query data and fetch configured metrics
        2. Check search results and metrics with provided condition
        3. Execute the action if condition is satisfied

        """
        current_api = singleton.active_api()
        try:
            dataset = current_api.datasets.find_by_name(self.dataset)
            self._LOGGER.debug(f"Found listener dataset {dataset.name}")
        except NotFoundApiError:
            self._LOGGER.warning(f"Not found dataset <{self.dataset}>")
            return

        ctx = RGListenerContext(
            listener=self,
            query_params=self.query_params,
            metrics=self.__compute_metrics__(current_api, dataset, query=self.formatted_query),
        )
        if self.condition is None:
            self._LOGGER.debug("No condition found! Running action...")
            return self.__run_action__(ctx, *args, **kwargs)

        search_results = current_api.search.search_records(
            name=self.dataset,
            task=dataset.task,
            size=0,
            query_text=self.formatted_query,
        )

        ctx.search = Search(
            total=search_results.total,
            query_params=copy.deepcopy(ctx.query_params),
        )
        condition_args = [ctx.search]
        if self.metrics:
            condition_args.append(ctx.metrics)

        self._LOGGER.debug(f"Evaluate condition with arguments: {condition_args}")
        if self.condition(*condition_args):
            self._LOGGER.debug("Condition passed! Running action...")
            return self.__run_action__(ctx, *args, **kwargs)

    def __compute_metrics__(self, current_api, dataset, query: str) -> Metrics:
        metrics = {}
        for metric in self.metrics:
            metrics.update(
                {
                    metric: current_api.metrics.metric_summary(
                        name=self.dataset,
                        task=dataset.task,
                        metric=metric,
                        query=query,
                    )
                }
            )
        return Metrics(**metrics)

    def __run_action__(self, ctx: Optional[RGListenerContext] = None, *args, **kwargs):
        try:
            action_args = [ctx] if ctx else []
            if self.query_records:
                action_args.insert(0, argilla_v1.load(name=self.dataset, query=self.formatted_query))
            self._LOGGER.debug(f"Running action with arguments: {action_args}")
            return self.action(*args, *action_args, **kwargs)
        except:  # noqa: E722
            import traceback

            print(traceback.format_exc())
            return schedule.CancelJob


def listener(
    dataset: str,
    query: Optional[str] = None,
    metrics: Optional[List[str]] = None,
    condition: Optional[ListenerCondition] = None,
    with_records: bool = True,
    execution_interval_in_seconds: int = 30,
    **query_params,
):
    """
    Configures the decorated function as an argilla listener.

    Args:
        dataset: The dataset name.
        query: The query string.
        metrics: Required metrics for listener condition.
        condition: Defines condition over search and metrics that launch action when is satisfied.
        with_records: Include records as part of action arguments. If ``False``,
            only the listener context ``RGListenerContext`` will be passed. Default: ``True``.
        execution_interval_in_seconds: Define the execution interval in seconds when listener
            iteration will be executed.
        **query_params: Dynamic parameters used in the query. These parameters will be available
            via the listener context and can be updated for subsequent queries.
    """

    def inner_decorator(func):
        return RGDatasetListener(
            dataset=dataset,
            action=func,
            condition=condition,
            query=query,
            query_params=query_params,
            metrics=metrics,
            query_records=with_records,
            interval_in_seconds=execution_interval_in_seconds,
        )

    return inner_decorator



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, List, Optional, Union

from argilla_v1.client.feedback.schemas.enums import QuestionTypes
from argilla_v1.pydantic_v1 import (
    BaseModel,
    StrictInt,
    StrictStr,
    confloat,
    conint,
    constr,
    parse_obj_as,
    root_validator,
)

if TYPE_CHECKING:
    from argilla_v1.client.feedback.schemas.questions import QuestionSchema


class RankingValueSchema(BaseModel):
    """Schema for the `RankingQuestion` response value for a `RankingQuestion`. Note that
    we may have more than one record in the same rank.

    Args:
        value: The value of the record.
        rank: The rank of the record.
    """

    value: StrictStr
    rank: Optional[conint(ge=1)] = None


class SpanValueSchema(BaseModel):
    """Schema for the `SpanQuestion` response value for a `SpanQuestion`.

    Args:
        label: The label value of the span.
        start: The start of the span.
        end: The end of the span.
        score: The score of the span.
    """

    label: constr(min_length=1)
    start: conint(ge=0)
    end: conint(ge=0)
    score: Optional[confloat(ge=0.0, le=1.0)] = None

    @root_validator
    def check_span(cls, values):
        if values["end"] <= values["start"]:
            raise ValueError("The end of the span must be greater than the start.")
        return values


ResponseValue = Union[
    StrictStr,
    StrictInt,
    List[str],
    List[dict],
    List[RankingValueSchema],
    List[SpanValueSchema],
]

RESPONSE_VALUE_FOR_QUESTION_TYPE = {
    QuestionTypes.text: str,
    QuestionTypes.label_selection: str,
    QuestionTypes.multi_label_selection: List[str],
    QuestionTypes.ranking: List[RankingValueSchema],
    QuestionTypes.rating: int,
    QuestionTypes.span: List[SpanValueSchema],
}


def parse_value_response_for_question(question: "QuestionSchema", value: ResponseValue) -> ResponseValue:
    question_type = question.type
    response_type = RESPONSE_VALUE_FOR_QUESTION_TYPE[question_type]

    if isinstance(value, (dict, list)):
        return parse_obj_as(response_type, value)
    elif not isinstance(value, response_type):
        raise ValueError(f"Value {value} is not valid for question type {question_type}. Expected {response_type}.")

    return value


def normalize_response_value(value: ResponseValue) -> ResponseValue:
    """Normalize the response value."""
    if not isinstance(value, list) or not all([isinstance(v, dict) for v in value]):
        return value

    new_value = []
    for v in value:
        if "start" in v and "end" in v:
            new_value.append(SpanValueSchema(**v))
        elif "value" in v:
            new_value.append(RankingValueSchema(**v))
        else:
            raise ValueError("Invalid value", value)
    return new_value



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from argilla_v1.client.api import Api
from argilla_v1.client.models import TextClassificationRecord
from argilla_v1.monitoring.base import BaseMonitor
from argilla_v1.monitoring.types import MissingType
from argilla_v1.pydantic_v1 import BaseModel

try:
    from transformers import (
        Pipeline,
        TextClassificationPipeline,
        ZeroShotClassificationPipeline,
    )
    from transformers import __version__ as _transformers_version
except ModuleNotFoundError:
    TextClassificationPipeline = MissingType
    Pipeline = MissingType
    ZeroShotClassificationPipeline = MissingType
    _transformers_version = None


class LabelPrediction(BaseModel):
    label: str
    score: float


class HuggingFaceMonitor(BaseMonitor):
    @staticmethod
    def fetch_transformers_version(config) -> Optional[str]:
        config_dict = config.to_dict()
        return config_dict.get("transformers_version", _transformers_version)

    @property
    def model_config(self):
        return self.__model__.model.config

    def _prepare_log_data(
        self,
        data: List[Tuple[str, Dict[str, Any], List[LabelPrediction]]],
        multi_label: bool = False,
    ) -> Dict[str, Any]:
        agent = self.model_config.name_or_path

        records = []
        for input_, metadata, predictions in data:
            record = TextClassificationRecord(
                text=input_ if isinstance(input_, str) else None,
                inputs=input_ if not isinstance(input_, str) else None,
                prediction=[(prediction.label, prediction.score) for prediction in predictions],
                prediction_agent=agent,
                metadata=metadata or {},
                multi_label=multi_label,
                event_timestamp=datetime.utcnow(),
            )
            records.append(record)

        dataset_name = self.dataset
        if multi_label:
            dataset_name += "_multi"

        return dict(
            records=records,
            name=dataset_name,
            tags={
                "name": self.model_config.name_or_path,
                "transformers_version": self.fetch_transformers_version(self.model_config),
                "model_type": self.model_config.model_type,
                "task": self.__model__.task,
            },
            metadata=self.model_config.to_dict(),
            verbose=False,
        )


class ZeroShotMonitor(HuggingFaceMonitor):
    def __call__(
        self,
        sequences: Union[str, List[str]],
        candidate_labels: List[str],
        *args,
        **kwargs,
    ):
        metadata = (kwargs.pop("metadata", None) or {}).copy()
        hypothesis_template = kwargs.get("hypothesis_template", "@default")
        multi_label = kwargs.get("multi_label", kwargs.get("multi_class", False))
        batch_predictions = self.__model__(sequences, candidate_labels, *args, **kwargs)
        try:
            if not isinstance(sequences, list):
                sequences = [sequences]

            predictions = batch_predictions
            if not isinstance(batch_predictions, list):
                predictions = [batch_predictions]

            predictions = [
                [
                    LabelPrediction(label=label, score=score)
                    for label, score in zip(prediction["labels"], prediction["scores"])
                ]
                for prediction in predictions
            ]

            if not metadata:
                metadata = [{}] * len(sequences)
            elif not isinstance(metadata, list):
                metadata = [metadata]
            for meta in metadata:
                meta.update(
                    {
                        "labels": candidate_labels,
                        "hypothesis_template": hypothesis_template,
                    }
                )
            filtered_data = [
                (input_, meta, predictions_)
                for input_, meta, predictions_ in zip(sequences, metadata, predictions)
                if self.is_record_accepted()
            ]
            if filtered_data:
                self.send_records(filtered_data, multi_label=multi_label)

        finally:
            return batch_predictions


class TextClassificationMonitor(HuggingFaceMonitor):
    """Configures monitoring over Hugging Face text classification pipelines"""

    def __call__(self, inputs, *args, **kwargs):
        metadata = kwargs.pop("metadata", None)
        batch_predictions = self.__model__(inputs, *args, **kwargs)
        try:
            if not isinstance(inputs, list):
                inputs = [inputs]
            if not metadata:
                metadata = [None] * len(inputs)
            elif not isinstance(metadata, list):
                metadata = [metadata]

            predictions_ = []
            for pred in batch_predictions:
                if isinstance(pred, dict):
                    pred = [pred]
                predictions_.append([LabelPrediction.parse_obj(p) for p in pred])

            filtered_data = [
                (input_, meta, predictions)
                for input_, meta, predictions in zip(inputs, metadata, predictions_)
                if self.is_record_accepted()
            ]
            if filtered_data:
                self.send_records(filtered_data)

        finally:
            return batch_predictions


def huggingface_monitor(
    pl: Pipeline, api: Api, dataset: str, sample_rate: float, log_interval: float
) -> Optional[Pipeline]:
    if isinstance(pl, TextClassificationPipeline):
        return TextClassificationMonitor(
            pl,
            api=api,
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
        )
    if isinstance(pl, ZeroShotClassificationPipeline):
        return ZeroShotMonitor(
            pl,
            api=api,
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
        )
    return None
